/*******************************************************************************
 * Copyright (C) 2005-2015 Alfresco Software Limited.
 *
 * This file is part of the Alfresco Mobile iOS App.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 ******************************************************************************/

#import "FileProvider.h"
#import <UIKit/UIKit.h>
#import "PersistentQueueStore.h"
#import "FileMetadata.h"
#import "SharedConstants.h"
#import "KeychainUtils.h"
#import "UserAccountWrapper.h"
#import "AlfrescoFileManager+Extensions.h"
#import "NSFileManager+Extension.h"
#import "Utilities.h"

static NSString * const kAccountsListIdentifier = @"AccountListNew";

@interface FileProvider ()
@property (nonatomic, strong) PersistentQueueStore *queueStore;
@property (nonatomic, strong) NSMutableDictionary *accountIdentifierToSessionMappings;
@end

@implementation FileProvider

- (instancetype)init
{
    self = [super init];
    if (self)
    {
        [self.fileCoordinator coordinateWritingItemAtURL:[self documentStorageURL] options:0 error:nil byAccessor:^(NSURL *newURL) {
            // ensure the documentStorageURL actually exists
            NSError *error = nil;
            [[NSFileManager defaultManager] createDirectoryAtURL:newURL withIntermediateDirectories:YES attributes:nil error:&error];
        }];
    }
    return self;
}

#pragma mark - Custom Getters and Setters

- (NSFileCoordinator *)fileCoordinator
{
    NSFileCoordinator *fileCoordinator = [[NSFileCoordinator alloc] init];
    [fileCoordinator setPurposeIdentifier:[self providerIdentifier]];
    return fileCoordinator;
}

- (PersistentQueueStore *)queueStore
{
    if (!_queueStore)
    {
        _queueStore = [[PersistentQueueStore alloc] initWithGroupContainerIdentifier:kSharedAppGroupIdentifier];
    }
    return _queueStore;
}

- (NSMutableDictionary *)accountIdentifierToSessionMappings
{
    if (!_accountIdentifierToSessionMappings)
    {
        _accountIdentifierToSessionMappings = [NSMutableDictionary dictionary];
    }
    return _accountIdentifierToSessionMappings;
}

#pragma mark - Private Methods

- (FileMetadata *)fileMetadataForURL:(NSURL *)fileURL
{
    FileMetadata *returnMetadata = nil;
    
    /*
     * We need the unique document identifier appended onto the URL.
     * Since we can't get access to the document, in order to generate this, we simply take the 
     * last two path components from the url provided in order to generate the search url.
     */
    NSArray *pathComponents = fileURL.pathComponents;
    NSArray *folderAndFilePathComponents = [pathComponents subarrayWithRange:NSMakeRange(pathComponents.count - 2, 2)];
    NSString *lastTwoPathString = [NSString pathWithComponents:folderAndFilePathComponents];
    
    NSString *lastPartDocumentStorageURL = self.documentStorageURL.lastPathComponent;
    if([lastTwoPathString containsString:lastPartDocumentStorageURL])
    {
        lastTwoPathString = folderAndFilePathComponents.lastObject;
    }
    
    NSURL *searchURL = [self.documentStorageURL URLByAppendingPathComponent:lastTwoPathString];
    NSPredicate *searchPredicate = [NSPredicate predicateWithFormat:@"fileURL == %@", searchURL];
    NSArray *urlSearchResultArray = [self.queueStore.queue filteredArrayUsingPredicate:searchPredicate];
    
    returnMetadata = urlSearchResultArray.lastObject;
    
    return returnMetadata;
}

- (UserAccountWrapper *)userAccountForMetadataItem:(FileMetadata *)metadata
{
    NSError *keychainError = nil;
    NSArray *accounts = [KeychainUtils savedAccountsForListIdentifier:kAccountsListIdentifier error:&keychainError];
    
    if (keychainError)
    {
        AlfrescoLogError(@"Error retreiving accounts. Error: %@", keychainError.localizedDescription);
    }
    
    // Get the account for the file
    NSPredicate *predicate = [NSPredicate predicateWithFormat:@"accountIdentifier == %@", metadata.accountIdentifier];
    NSArray *accountArray = [accounts filteredArrayUsingPredicate:predicate];
    UserAccount *keychainAccount = accountArray.firstObject;
    UserAccountWrapper *account = [[UserAccountWrapper alloc] initWithUserAccount:keychainAccount];
    account.selectedNetworkIdentifier = metadata.networkIdentifier;
    
    return account;
}

- (void)loginToAccount:(id<AKUserAccount>)account completionBlock:(void (^)(BOOL successful, id<AlfrescoSession> session, NSError *loginError))completionBlock
{
    AKLoginService *loginService = [[AKLoginService alloc] init];
    [loginService loginToAccount:account networkIdentifier:account.selectedNetworkIdentifier completionBlock:^(BOOL successful, id<AlfrescoSession> session, NSError *loginError) {
        if (successful)
        {
            self.accountIdentifierToSessionMappings[account.identifier] = session;
        }
        
        completionBlock(successful, session, loginError);
    }];
}

- (void)uploadDocument:(AlfrescoDocument *)document sourceURL:(NSURL *)fileURL session:(id<AlfrescoSession>)session completionBlock:(void (^)(AlfrescoDocument *document, NSError *updateError))completionBlock
{
    AlfrescoContentFile *contentFile = [[AlfrescoContentFile alloc] initWithUrl:fileURL];
    
    AlfrescoVersionService *versionService = [[AlfrescoVersionService alloc] initWithSession:session];
    [versionService checkoutDocument:document completionBlock:^(AlfrescoDocument *checkoutDocument, NSError *checkoutError) {
        if (checkoutError)
        {
            if(checkoutError.code != kAlfrescoErrorCodeVersion)
            {
                completionBlock(checkoutDocument, checkoutError);
            }
        }
        else
        {
            [versionService checkinDocument:document asMajorVersion:NO contentFile:contentFile properties:nil comment:nil completionBlock:^(AlfrescoDocument *checkinDocument, NSError *checkinError) {
                completionBlock(checkinDocument, checkinError);
            } progressBlock:nil];
        }
    }];
}

- (void)saveDocumentAtURL:(NSURL *)readingURL toURL:(NSURL *)writingURL overwritingExistingFile:(BOOL)shouldOverwrite
{
    NSError *copyError = nil;
    NSFileManager *fileManager = [[NSFileManager alloc] init];
    if(!shouldOverwrite && [fileManager fileExistsAtPath:[writingURL path]])
    {
        NSString *filename = [self fileNameAppendedWithDate:writingURL.lastPathComponent];
        writingURL = [[writingURL URLByDeletingLastPathComponent] URLByAppendingPathComponent:filename];
    }
    [fileManager copyItemAtURL:readingURL toURL:writingURL overwritingExistingFile:shouldOverwrite error:&copyError];
    
    if (copyError)
    {
        AlfrescoLogError(@"Unable to copy file at path: %@, to location: %@. Error: %@", readingURL, writingURL, copyError.localizedDescription);
    }
}

- (NSString *)fileNameAppendedWithDate:(NSString *)name
{
    NSDateFormatter* dateFormatter = [[NSDateFormatter alloc] init];
    [dateFormatter setDateFormat:@"yyyy-MM-dd-HH-mm-ss"];
    NSString *dateString = [dateFormatter stringFromDate:[NSDate date]];
    NSString *fileExtension = name.pathExtension;
    NSString *fileName = [NSString stringWithFormat:@"%@_%@", name.stringByDeletingPathExtension, dateString];
    
    if (fileExtension.length > 0)
    {
        fileName = [fileName stringByAppendingPathExtension:fileExtension];
    }
    
    return fileName;
}

#pragma mark - File Provider Methods

- (void)providePlaceholderAtURL:(NSURL *)url completionHandler:(void (^)(NSError *error))completionHandler
{
    FileMetadata *fileMetadata = [self fileMetadataForURL:url];
    AlfrescoDocument *document = (AlfrescoDocument *)fileMetadata.repositoryNode;
    NSString *fileName = document.name;
    unsigned long long fileSize = document.contentLength;
    
    NSURL *placeholderURL = [NSFileProviderExtension placeholderURLForURL:[self.documentStorageURL URLByAppendingPathComponent:fileName]];
    
    NSError *placeholderWriteError = nil;
    [self.fileCoordinator coordinateWritingItemAtURL:placeholderURL options:0 error:&placeholderWriteError byAccessor:^(NSURL *newURL) {
        NSDictionary *metadata = @{NSURLNameKey : document.name, NSURLFileSizeKey : @(fileSize), NSURLContentModificationDateKey : document.modifiedAt};
        [NSFileProviderExtension writePlaceholderAtURL:placeholderURL withMetadata:metadata error:NULL];
    }];
    
    if (completionHandler)
    {
        completionHandler(placeholderWriteError);
    }
}

- (void)startProvidingItemAtURL:(NSURL *)url completionHandler:(void (^)(NSError *))completionHandler
{
    NSFileManager *fileManager = [[NSFileManager alloc] init];
    NSString *pathToFile = url.path;
    
    if ([fileManager fileExistsAtPath:pathToFile])
    {
        completionHandler(nil);
    }
    else
    {
        FileMetadata *metadata = [self fileMetadataForURL:url];
        
        if (metadata.saveLocation == FileMetadataSaveLocationRepository)
        {
            UserAccountWrapper *account = [self userAccountForMetadataItem:metadata];
            [self loginToAccount:account completionBlock:^(BOOL successful, id<AlfrescoSession> session, NSError *loginError) {
                if (successful)
                {
                    NSOutputStream *outputStream = [NSOutputStream outputStreamWithURL:url append:NO];
                    
                    AlfrescoDocumentFolderService *docService = [[AlfrescoDocumentFolderService alloc] initWithSession:session];
                    [docService retrieveContentOfDocument:(AlfrescoDocument *)metadata.repositoryNode outputStream:outputStream completionBlock:^(BOOL succeeded, NSError *error) {
                        if (error)
                        {
                            completionHandler(error);
                        }
                        else
                        {
                            completionHandler(nil);
                        }
                    } progressBlock:nil];
                }
                else
                {
                    completionHandler(loginError);
                }
            }];
        }
        else if (metadata.saveLocation == FileMetadataSaveLocationLocalFiles)
        {
            NSString *downloadContentPath = [[AlfrescoFileManager sharedManager] downloadsContentFolderPath];
            NSString *fullSourcePath = [downloadContentPath stringByAppendingPathComponent:url.lastPathComponent];
            NSURL *sourceURL = [NSURL fileURLWithPath:fullSourcePath];
            
            NSError *copyError = nil;
            NSFileManager *fileManager = [[NSFileManager alloc] init];
            [fileManager copyItemAtURL:sourceURL toURL:url error:&copyError];
            
            if (copyError)
            {
                AlfrescoLogError(@"Unable to copy item from: %@, to: %@. Error: %@", sourceURL, url, copyError.localizedDescription);
            }
            
            completionHandler(copyError);
        }
    }
}

- (void)itemChangedAtURL:(NSURL *)url
{
    AlfrescoLogInfo(@"Item changed at URL %@", url);
    
    NSFileManager *fileManager = [[NSFileManager alloc] init];
    NSDictionary *fileAttributes = [fileManager attributesOfItemAtPath:url.path error:nil];
    NSDate *urlModificationDate = fileAttributes[NSFileModificationDate];
    
    FileMetadata *metadata = [self fileMetadataForURL:url];
    if(metadata)
    {
        // the document exists in the AlfrescoRepository
        AlfrescoDocument *repoNode = (AlfrescoDocument *)metadata.repositoryNode;
        
        if (![repoNode.modifiedAt isEqualToDate:urlModificationDate] && metadata.status != FileMetadataStatusUploading)
        {
            if (metadata.saveLocation == FileMetadataSaveLocationRepository)
            {
                UserAccountWrapper *account = [self userAccountForMetadataItem:metadata];
                
                // Coordinate the reading of the file for uploading
                [self.fileCoordinator coordinateReadingItemAtURL:metadata.fileURL options:NSFileCoordinatorReadingForUploading error:nil byAccessor:^(NSURL *newReadingURL) {
                    // define an upload block
                    __block BOOL networkOperationCallbackComplete = NO;
                    void (^uploadBlock)(id<AlfrescoSession>session) = ^(id<AlfrescoSession>session) {
                        AlfrescoDocument *updateDocument = (AlfrescoDocument *)metadata.repositoryNode;
                        [self uploadDocument:updateDocument sourceURL:newReadingURL session:session completionBlock:^(AlfrescoDocument *document, NSError *updateError) {
                            if (updateError)
                            {
                                AlfrescoLogError(@"Error Updating Document: %@. Error: %@", updateDocument.name, updateError.localizedDescription);
                                NSString *downloadContentPath = [[AlfrescoFileManager sharedManager] downloadsContentFolderPath];
                                NSString *filename = url.lastPathComponent;
                                filename = [Utilities filenameWithoutVersionFromFilename:filename nodeIdentifier:repoNode.identifier];
                                NSString *fullDestinationPath = [downloadContentPath stringByAppendingPathComponent:filename];
                                NSURL *destinationURL = [NSURL fileURLWithPath:fullDestinationPath];
                                [self.fileCoordinator coordinateWritingItemAtURL:destinationURL options:NSFileCoordinatorWritingForReplacing error:nil byAccessor:^(NSURL * _Nonnull newURL) {
                                    [self saveDocumentAtURL:newReadingURL toURL:newURL overwritingExistingFile:YES];
                                }];
                            }
                            else
                            {
                                AlfrescoLogInfo(@"Successfully updated document: %@, Modified At: %@, Created At: %@", document.name, document.modifiedAt, document.createdAt);
                                metadata.repositoryNode = document;
                            }
                            networkOperationCallbackComplete = YES;
                            metadata.status = FileMetadataStatusPendingUpload;
                            [self.queueStore saveQueue];
                        }];
                        
                        // Set the metadata to uploading - ensure the upload doesnt start again.
                        metadata.status = FileMetadataStatusUploading;
                        [self.queueStore saveQueue];
                    };
                    
                    // Session exists, use that, else do a login and then upload.
                    id<AlfrescoSession> cachedSession = self.accountIdentifierToSessionMappings[account.identifier];
                    if (!cachedSession)
                    {
                        [self loginToAccount:account completionBlock:^(BOOL successful, id<AlfrescoSession> session, NSError *loginError) {
                            if (successful)
                            {
                                uploadBlock(session);
                            }
                            else
                            {
                                AlfrescoLogError(@"Error Logging In: %@", loginError.localizedDescription);
                                NSString *downloadContentPath = [[AlfrescoFileManager sharedManager] downloadsContentFolderPath];
                                NSString *filename = url.lastPathComponent;
                                filename = [Utilities filenameWithoutVersionFromFilename:filename nodeIdentifier:repoNode.identifier];
                                NSString *fullDestinationPath = [downloadContentPath stringByAppendingPathComponent:filename];
                                NSURL *destinationURL = [NSURL fileURLWithPath:fullDestinationPath];
                                [self.fileCoordinator coordinateWritingItemAtURL:destinationURL options:NSFileCoordinatorWritingForReplacing error:nil byAccessor:^(NSURL * _Nonnull newURL) {
                                    [self saveDocumentAtURL:newReadingURL toURL:newURL overwritingExistingFile:YES];
                                }];
                            }
                        }];
                        
                        /*
                         * Keep this object around long enough for the network operations to complete.
                         * Running as a background thread, seperate from the UI, so should not cause
                         * Any issues when blocking the thread.
                         */
                        do
                        {
                            [[NSRunLoop currentRunLoop] runMode:NSDefaultRunLoopMode beforeDate:[NSDate distantFuture]];
                        }
                        while (networkOperationCallbackComplete == NO);
                    }
                    else
                    {
                        uploadBlock(cachedSession);
                    }
                }];
                
            }
            else if (metadata.saveLocation == FileMetadataSaveLocationLocalFiles)
            {
                NSString *downloadContentPath = [[AlfrescoFileManager sharedManager] downloadsContentFolderPath];
                NSString *fullDestinationPath = [downloadContentPath stringByAppendingPathComponent:url.lastPathComponent];
                NSURL *destinationURL = [NSURL fileURLWithPath:fullDestinationPath];
                [self.fileCoordinator coordinateReadingItemAtURL:url options:NSFileCoordinatorReadingForUploading writingItemAtURL:destinationURL options:NSFileCoordinatorWritingForReplacing error:nil byAccessor:^(NSURL *newReadingURL, NSURL *newWritingURL) {
                    [self saveDocumentAtURL:newReadingURL toURL:newWritingURL overwritingExistingFile:YES];
                }];
            }
        }
    }
}

- (void)stopProvidingItemAtURL:(NSURL *)url
{
    FileMetadata *completedMetadata = [self fileMetadataForURL:url];
    [self.queueStore removeObjectFromQueue:completedMetadata];
    [self.queueStore saveQueue];
    
    [self.fileCoordinator coordinateWritingItemAtURL:url options:NSFileCoordinatorWritingForDeleting error:NULL byAccessor:^(NSURL *newURL) {
        [[NSFileManager defaultManager] removeItemAtURL:newURL error:NULL];
    }];
    [self providePlaceholderAtURL:url completionHandler:^(NSError *error) {}];
}

@end
